/***************************************************************************
 *   Copyright (C) 2010 by Stefan Baur                                     *
 *   kwatersaver@c0n.de                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA            *
 ***************************************************************************/

#include "kglwatersaver4.h"
#include "settings.h"
#include <QApplication>
#include <QDesktopWidget>
#include <QImage>

#include <QDebug>

//#define USE_X11_SCREENSHOT
#ifdef USE_X11_SCREENSHOT
#include <QX11Info>
#if defined Q_WS_X11 && !defined K_WS_QTONLY
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif
#endif


KGLWaterSaver4::KGLWaterSaver4(QWidget *parent)
: QGLWidget(parent)
{
	qDebug()<<"KGLWaterSaver4()";
	setAttribute( Qt::WA_NoSystemBackground );
	settings.load();
}

KGLWaterSaver4::~KGLWaterSaver4()
{
	qDebug()<<"~KGLWaterSaver4()";
	water.quit();
}

void KGLWaterSaver4::start()
{
	qDebug()<<"KGLWaterSaver4::start()";
	m_timerId = startTimer(1000/settings.fps);
}

void KGLWaterSaver4::stop()
{
	qDebug()<<"KGLWaterSaver4::stop()";
	killTimer( m_timerId );
}


unsigned int loadQImageAsTexture(QImage bgImage)
{
	int imgSize,maxTexture,maxDesktop,maxImage;
	unsigned int textureID;

	if(bgImage.isNull()) //FIXME: whoo ... we do not have a image!?
		{
		QImage errImage(128,128,QImage::Format_RGB32);
		QPainter errPaint(&errImage);
		
		errPaint.fillRect(0,0,128,128,Qt::black);
		errPaint.setBrush(Qt::red);
		errPaint.setPen(Qt::red);
		errPaint.drawText(40,16,"failed to");
		errPaint.drawText(52,48,"load");
		errPaint.drawText(40,80,"image");
		errPaint.drawText(60,112,"...");
		bgImage=errImage;
		}

	glGetIntegerv(GL_MAX_TEXTURE_SIZE,&maxTexture);
	if(maxTexture<64)maxTexture=64; //we want at least 64x64 textures as it makes no sense to have less

	maxDesktop=QApplication::desktop()->width();
	if(QApplication::desktop()->height()>maxDesktop)maxDesktop=QApplication::desktop()->height();

	maxImage=bgImage.width();
	if(bgImage.height()>maxImage)maxImage=bgImage.height();

	qDebug(" maxTexture: %i, maxImage: %i, maxDesktop: %i",maxTexture,maxImage,maxDesktop);

	if(!strstr((char*)glGetString(GL_EXTENSIONS),(char*)"GL_ARB_texture_non_power_of_two") || settings.usePowerOf2Textures)
	{
			// increase imgSize by power of 2 to find the best fitting texture resolution for current image
		for(imgSize=64;imgSize<maxTexture;imgSize*=2)
		{
				//are we already as big as the image? Then break!
			if( imgSize>=maxImage )
				break;

				//are we already near the desktop resolution? Then break!
			if( imgSize*1.5>maxDesktop ) //*2 ?
				break;
		}

				// imgSize should now be a power of two - big enough for the image and desktop
		if(imgSize>maxTexture)imgSize=maxTexture;
		qDebug(" using %i x %i Texture!",imgSize,imgSize);

		bgImage=bgImage.scaled(imgSize,imgSize);
	}
	else // use non power of two textures
	{
		imgSize=maxImage;
		if(imgSize>maxDesktop)
			imgSize=maxDesktop; //we do not need textures bigger than the screen!
		if(imgSize>maxTexture)
			imgSize=maxTexture; //we need to be smaller than max texture size!
		
		bgImage=bgImage.scaled(bgImage.width()>imgSize?imgSize:bgImage.width(),bgImage.height()>imgSize?imgSize:bgImage.height());
		qDebug(" using %i x %i Texture!",bgImage.width(),bgImage.height());
	}

		//make it OpenGL compatible
	bgImage=QGLWidget::convertToGLFormat(bgImage);

		//and load it into the graphics card
	glGenTextures(1,&textureID);
	glBindTexture(GL_TEXTURE_2D,textureID);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR );
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR );
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP); //GL_CLAMP or GL_REPEAT
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glEnable(GL_TEXTURE_2D);
	glTexImage2D(GL_TEXTURE_2D,0,3,bgImage.width(),bgImage.height(),0,GL_RGBA,GL_UNSIGNED_BYTE,bgImage.bits() );
	return textureID;
}

void KGLWaterSaver4::reload()
{
	//NOTE: if water.Settings.bgTexID already is bound water will release it
	water.quit();
	water.init();
	
	if(settings.bgSource=="desktop")
	{
		water.Settings.bgTexID=loadQImageAsTexture(grabRootWindow());
	}
	else
	{
		water.Settings.bgTexID=loadQImageAsTexture(QImage(settings.bgSource));
	}
	
}




QImage KGLWaterSaver4::grabRootWindow()
{
#ifdef USE_X11_SCREENSHOT
	Display *dsp = QX11Info::display();
	Window rootwin = RootWindow( dsp, QX11Info::appScreen() );
	
	static XImage *xRootWin=0;

	// grab contents of root window
	if( xRootWin )
		XDestroyImage( xRootWin );

	xRootWin = XGetImage( dsp, rootwin, 0, 0, width(),
	                      height(), AllPlanes, ZPixmap);
	if(xRootWin)
	{
		return QImage((const uchar*)xRootWin->data,xRootWin->width,xRootWin->height,xRootWin->bytes_per_line, QImage::Format_RGB32);
	}
	else
		qDebug()<<"KGLWaterSaver4::grabRootWindow: failed to grab rootWindow";
	
	QImage errImage(128,128,QImage::Format_RGB32);
	QPainter errPaint(&errImage);
	
	errPaint.fillRect(0,0,128,128,Qt::black);
	errPaint.setBrush(Qt::red);
	errPaint.setPen(Qt::red);
	errPaint.drawText(40,16,"failed to");
	errPaint.drawText(52,48,"grab");
	errPaint.drawText(40,80,"desktop");
	errPaint.drawText(60,112,"...");
	return errImage;
#else
	return QPixmap::grabWindow(QApplication::desktop()->winId()).toImage();
#endif
}


void KGLWaterSaver4::initializeGL()
{
	qDebug()<<"KGLWaterSaver4::initializeGL()";
	qglClearColor(QColor(Qt::black));

	reload();

	QImage envImage(":/res/envmap.jpg");//(128,128,QImage::Format_RGB32);
	water.Settings.envTexID=loadQImageAsTexture(envImage);

	t.start();
}

void KGLWaterSaver4::resizeGL(int width, int height)
{
	qDebug()<<"KGLWaterSaver4::resizeGL("<<width<<","<<height<<")";
		//set up for fine 2d rendering in the area [ -1/-1 , 1,1 ]

	glViewport( 0, 0, width, height );

	glMatrixMode(GL_PROJECTION);
	 glLoadIdentity();
	 glOrtho( -1, 1, -1, 1, 5.0, 60.0 );

	glMatrixMode(GL_MODELVIEW);
	 glLoadIdentity();
}



void KGLWaterSaver4::paintGL()
{
//	qDebug()<<"KGLWaterSaver4::paintGL()";
	//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	water.render();
}

void KGLWaterSaver4::timerEvent(QTimerEvent * )
{
//	qDebug()<<"KGLWaterSaver4::timerEvent()";
	if(water.update(t.elapsed() / 1000.0))
		updateGL();
}

